#include <Arduino.h>
#include <ArduinoLog.h>
#include "config.h"
#include <Pins.h>
#include <NFC.h>
#include <OTAProxy.h>
#include <helpers.h>
#include <Display.h>

#include <WiFi.h>
#include <WebServer.h>

#include <WiFiClient.h>
#include <PubSubClient.h>

WiFiClient espClient;

PubSubClient *mqtt;
NFC *nfc;
OTAProxy *ota;
Display *display;

unsigned long otatimeout = 3000;
unsigned long lastotatime;

void setup_wifi()
{
    delay(10);
    Log.trace(F("Connecting Wifi ..."));

    WiFi.mode(WIFI_STA);
    WiFi.begin(WLAN_SSID, WLAN_PASS);

    while (WiFi.status() != WL_CONNECTED)
    {
        delay(500);
    }

    randomSeed(micros());
    Log.trace(F("WiFi connected"));
}

void reconnect()
{
    while (!mqtt->connected())
    {
        String clientId = "FabReader_";
        clientId += String(FABREADERID);

        Log.trace(F("Connecting MQTT ..."));
        bool connected = false;
        if (strcmp(MQTT_USERNAME, "") == 0)
        {
            connected = mqtt->connect(clientId.c_str());
        }
        else
        {
            connected = mqtt->connect(clientId.c_str(), MQTT_USERNAME, MQTT_PASSWORD);
        }

        if (connected)
        {
            Log.trace(F("MQTT connected"));
            char id[6] = "00000";
            sprintf(id, "%05d", FABREADERID);
            mqtt->publish("fabreader", id);

            char topic_requestOTA[] = "fabreader/00000/requestOTA";
            sprintf(topic_requestOTA, "fabreader/%05d/requestOTA", FABREADERID);
            mqtt->subscribe(topic_requestOTA);

            char topic_stopOTA[] = "fabreader/00000/stopOTA";
            sprintf(topic_stopOTA, "fabreader/%05d/stopOTA", FABREADERID);
            mqtt->subscribe(topic_stopOTA);

            char topic_display[] = "fabreader/00000/display/#";
            sprintf(topic_display, "fabreader/%05d/display/#", FABREADERID);
            mqtt->subscribe(topic_display);

            display->writeTitle("Connected");
            display->writeInfo("");
        }
        else
        {
            display->writeTitle("Reconnect");
            display->writeInfo("MQTT");

            Log.error(F("MQTT connection failed, rc=%d, try again in 5 seconds"), mqtt->state());
            delay(5000);
        }
    }
}

void callback(char *topic, byte *payload, unsigned int length)
{
    Log.trace(F("Receive Message: %s"), topic);
    if (ota->hasActiveOTA())
    {
        ota->continueOTA(topic, payload, length);
    }
    display->updateByMQTT(topic, payload, length);
}

void setup()
{
    Serial.begin(115200);

    Log.begin(LOG_LEVEL_VERBOSE, &Serial); // Initialize ArduinoLog

    // Optional: add a timestamp to each log entry.
    // Log.begin(LOG_LEVEL_VERBOSE, &Serial, true);

    Log.trace(F("\n\n\n"));
    Log.trace(F("Booting ..."));

    pinMode(PIN_BUZZER, OUTPUT);

    display = new Display(PIN_SDA, PIN_SCL, FABREADERID);
    display->clearReaderInfo();

    display->writeInfo("Start NFC ...");
    Log.trace(F("Connecting NFC ..."));

    nfc = new NFC(PIN_SDA, PIN_SCL);

    Log.trace(F("NFC connected"));

    display->writeInfo("Start WIFI ...");
    setup_wifi();

    display->writeInfo("Start MQTT ...");
    mqtt = new PubSubClient(espClient);
    mqtt->setServer(MQTT_BROKER, 1883);
    mqtt->setCallback(callback);

    display->writeInfo("Start OTA ...");
    ota = new OTAProxy(mqtt, nfc, FABREADERID);
}

void loop()
{
    if (!mqtt->connected())
    {
        reconnect();
    }
    mqtt->loop();

    if (!ota->hasActiveOTA())
    {
        if (nfc->checkforCard())
        {
            Log.trace(F("Card detected"));
            if (nfc->connecttoCard())
            {
                Log.trace(F("Card connected"));
                lastotatime = millis();

                display->createReaderInfo("Run OTA");
                ota->startOTA();
            }
            else
            {
                display->createReaderInfo("Retry Card");
            }
        }
    }
    if (ota->hasActiveOTA())
    {
        if (millis() - lastotatime > otatimeout)
        {
            ota->cancelOTA();
            display->clearReaderInfo();
        }
    }
}